#if !defined(__AGGREGATOR_H)
    #define __AGGREGATOR_H
/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : aggregator.h                                             */
/*                                                                            */
/* Description     : This header file contains function prototypes and        */
/*                   constants for aggregator.c                               */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                      Date: Apr 2000                     */
/*                                                                            */
/******************************************************************************/

/* hack to get build to work ok under NT 4.0 */
#if 0  /* no longer seems to be required??? */
	#if !defined(_WIN32_WINNT)
		/* hack to get build to work ok under NT 4.0 */
		#define _WIN32_WINNT 0x0400
	#endif
#endif

#if defined(WIN32)
	#include <windows.h>
	#include <winbase.h>
#endif

#include <wchar.h>
#include <BipCci.h>	  /* Data types for Plug-in Node functions     */

#include <trace_publ.h>
#include <plugin.h>

#if defined(__THREADED)
    #include <threads.h>
#endif


#include <cmqc.h>
#include <unicode.h>
#include <pluginMsg.h>
#include <rc.h>

/* constants */
#define MAX_TIMER_NAME_LEN         256
#if !defined(IN)
    #define IN
#endif
#if !defined(OUT)
    #define OUT
#endif
#if !defined(INOUT)
    #define INOUT
#endif
#if !defined(TRUE)
    #define TRUE 1
#endif
#if !defined(FALSE)
    #define FALSE 0
#endif
#if !defined(bool_t)
    #define bool_t int
#endif

#define TIMED_OUT_QUEUE_NAME "SYSTEM.BROKER.TIMER.QUEUE"

#define PLUGIN_MESSAGE_CATALOGUE L"MQSeriesIntegratorPlugins"

#define AGGREGATOR_DATABASE L"AGGRDB"

#define AGGREGATE_CREATE_NODE L"MJBAggregateCreateNode"
#define AGGREGATE_CREATE_NODE_EXTERNAL_NAME_MB "SpAggregateCreate"

#define AGGREGATE_REPLY_NODE L"MJBARINode"
#define AGGREGATE_REPLY_NODE_EXTERNAL_NAME_MB "SpAggregateReply"

#define ATTR_AGGREGATE_NAME L"AggregateName"
#define ATTR_SEQUENCE_NUMBER L"SequenceNumber"
#define ATTR_REPLY_TIMEOUT_SECONDS L"ReplyTimeoutSeconds"
#define ATTR_REPLY_TO_QUEUE L"ReplyToQueue"
#define ATTR_SEPARATOR_STRING L"SeparatorString"
#define ATTR_HEX_SEPARATOR_STRING L"HexSeparatorString"
#define ATTR_AGGREGATE_REPLY_DESTINATION L"AggregatedReplyDestination"
#define ATTR_SET_TIMEOUT_CORREL_ID L"SetTimeoutCorrelId"
#define ATTR_PERSISTENT L"Recoverable"

#define TIMER_THREAD_ID L"AggregatorTimerThread       "
#define TIMER_THREAD_ID_MB "AggregatorTimerThread       "
#define AGG_TIMER_ID L"Agg-"
#define AGG_TIMER_ID_MB "Agg-"

/* return sequence numbers */
#define RSN_001     1
#define RSN_002     2
#define RSN_003     3
#define RSN_004     4
#define RSN_005     5
#define RSN_006     6
#define RSN_007     7
#define RSN_008     8
#define RSN_009     9
#define RSN_010     10
#define RSN_011     11
#define RSN_012     12
#define RSN_013     13
#define RSN_014     14
#define RSN_015     15
#define RSN_016     16
#define RSN_017     17
#define RSN_018     18
#define RSN_019     19
#define RSN_020     20
#define RSN_021     21
#define RSN_022     22
#define RSN_023     23
#define RSN_024     24
#define RSN_025     25
#define RSN_026     26
#define RSN_027     27
#define RSN_028     28
#define RSN_029     29
#define RSN_030     30
#define RSN_031     31
#define RSN_032     32
#define RSN_033     33
#define RSN_034     34
#define RSN_035     35
#define RSN_036     36
#define RSN_037     37
#define RSN_038     38
#define RSN_039     39
#define RSN_040     40
#define RSN_041     41
#define RSN_042     42
#define RSN_043     43
#define RSN_044     44
#define RSN_045     45
#define RSN_046     46
#define RSN_047     47
#define RSN_048     48
#define RSN_049     49
#define RSN_050     50
#define RSN_051     51
#define RSN_052     52
#define RSN_053     53
#define RSN_054     54
#define RSN_055     55
#define RSN_056     56
#define RSN_057     57
#define RSN_058     58
#define RSN_059     59
#define RSN_060     60
#define RSN_061     61
#define RSN_062     62
#define RSN_063     63
#define RSN_064     64
#define RSN_065     65
#define RSN_066     66
#define RSN_067     67
#define RSN_068     68
#define RSN_069     69
#define RSN_070     70

/* typedefs */
#include <time.h>

/* 
 * NB!!!! - any changes to the following structure should also be reflected in
 * the aggregateMemberLoggedData_t and the aggregateReplyLoggedData_t structures
 * defined in log.h
 */
typedef struct aggregateMember {
	int sequenceNumber,
		messageLength;
	time_t requestArrivalTime,
		   replyArrivalTime;
	char *mqmdP,
		 *messageP;
	CciChar replyToQueue[MQ_Q_NAME_LENGTH+1],
		    replyToQueueManager[MQ_Q_MGR_NAME_LENGTH+1];
	char originalMessageId[MQ_MSG_ID_LENGTH],
	     newMessageId[MQ_MSG_ID_LENGTH];
	struct aggregateInstance *aggregateInstanceP;
	struct msgIdIndexElement *indexP;
	struct aggregateMember *nextP,
		                   *prevP;
} aggregateMember_t;
#define AGGREGATE_MEMBER_DEFAULT 0,    \
	                             0,    \
	                             0,    \
	                             0,    \
								 NULL, \
								 NULL, \
								 L"",   \
								 L"",   \
								 "",   \
								 "",   \
								 NULL, \
								 NULL, \
								 NULL, \
								 NULL

typedef struct aggregateInstance {
	char aggregateId[MQ_CORREL_ID_LENGTH],
	     initialMsgId[MQ_MSG_ID_LENGTH];
	int memberCount;
	int outstandingReplies;
	time_t timeoutTime;
	bool_t timedOut,
		   commonMsgId;
	aggregateMember_t memberAnchor;
	struct deployedAggregate *deployedAggregateP;
	struct aggregateInstance *nextP,
		                       *prevP,
	                           *nextAlarmP,
		                       *prevAlarmP;
} aggregateInstance_t;
#define AGGREGATE_INSTANCE_DEFAULT   "",                        \
	                                 "",                        \
	                                 0,                         \
									 0,                         \
									 0,                         \
									 FALSE,                     \
									 FALSE,                     \
									{AGGREGATE_MEMBER_DEFAULT}, \
                                     NULL,                      \
                                     NULL,                      \
                                     NULL,                      \
                                     NULL,                      \
									 NULL

typedef struct deployedAggregate {
	CciChar *name;
	int timeoutInterval,
		aggregateReplyCount,
		aggregateCreateCount;
	bool_t outputToMQReply,
		   outputToDestinationList,
		   separatorStringSet,
		   persistent,
		   recoveryComplete,
		   aggregateCreateAndReplyExist,
		   setTimeoutCorrelId;
	CciChar replyToQueueName[MQ_Q_NAME_LENGTH + 1];
	char *separatorString;
	int  separatorStringLength;
	aggregateInstance_t aggregateAnchor;
	struct context *contextP;
	struct messageFlow *messageFlowEntryP;
	struct deployedAggregate *nextP,
		                     *prevP;
} deployedAggregate_t;
#define DEPLOYED_AGGREGATE_DEFAULT L"",                       \
	                               0,                         \
	                               0,                         \
	                               0,                         \
                                   FALSE,                     \
                                   FALSE,                     \
                                   FALSE,                     \
                                   FALSE,                     \
                                   FALSE,                     \
								   FALSE,                     \
								   FALSE,                     \
								   L"",                       \
								   NULL,                      \
                                   0,                         \
                                   {AGGREGATE_INSTANCE_DEFAULT},\
                                   NULL,                      \
                                   NULL,                      \
                                   NULL,                      \
								   NULL

typedef struct messageFlow {
	CciChar *name;
    deployedAggregate_t deployedAggregateAnchor;
	struct messageFlow *nextP,
		               *prevP;
} messageFlow_t;
#define MESSAGE_FLOW_DEFAULT L"",                         \
                             {DEPLOYED_AGGREGATE_DEFAULT},\
                             NULL,                        \
						     NULL

typedef enum
{
	AGGREGATE_CREATE = 1,
	AGGREGATE_REPLY
} aggregateType_t;

typedef enum
{
	MESSAGE_ID,
	CORREL_ID,
	REPLY_TO_QUEUE,
	REPLY_TO_QUEUE_MANAGER,
	MQMD_ELEMENT,
	BODY,
	DESTINATION,
	MQ_DESTINATION_LIST,
	DESTINATION_DATA,
	PUT_APPL_NAME
} matchType_t;

typedef struct trace_threadControlBlock
{
    int     indentationLevel,
	        myThreadId;
	bool_t  traceFileOpen;
	struct trace_threadControlBlock *nextP,
	                                *prevP;
} trace_threadControlBlock_t;

typedef struct msgIdIndexElement {
	char msgId[MQ_MSG_ID_LENGTH];
	aggregateMember_t *memberP;
	struct msgIdIndexElement *nextP,
		                     *prevP;
} msgIdIndexElement_t;
#define MESSAGE_ID_INDEX_DEFAULT "",   \
                                 NULL, \
                                 NULL, \
							     NULL

#include <log.h>

extern char *traceRetcodeStrings[];
extern int traceMaxRetcodeValue;

typedef struct addressList {
	int numAddresses,
		nextAddress;
	void *address[1];
} addressList_t;


#if defined(GLOBAL_DATA)
	/* global variables */
addressList_t *addrListP = NULL;
    threadMutex_t aggregateChainMutex,
				  alarmThreadSyncMutex,
				  logThreadSyncMutex,
	  			  logThreadWorkMutex,
	  			  uuidGenMutex;
	threadCondVar_t alarmThreadWorkCondVar,
		            alarmThreadSyncCondVar,
		            logThreadWorkCondVar,
		            logThreadSyncCondVar;
	bool_t shutdownRequested = FALSE,
		   alarmThreadInitialised = FALSE,
		   alarmThreadCreated = FALSE,
           logThreadInitialised = FALSE,
           logThreadCreated = FALSE,
		   brokerInfoObtained = FALSE,
		   multipleExecutionGroups = FALSE,
		   alarmThreadShuttingDown = FALSE,
		   logThreadShuttingDown = FALSE,
		   firstTraceCall = TRUE;

	messageFlow_t messageFlowAnchor = {MESSAGE_FLOW_DEFAULT};
    deployedAggregate_t deployedAggregateAnchor = {DEPLOYED_AGGREGATE_DEFAULT},
		                *deployedAggregateAnchorP = &deployedAggregateAnchor;
	aggregateInstance_t alarmAnchor = {AGGREGATE_INSTANCE_DEFAULT},
		                  timedOutAnchor = {AGGREGATE_INSTANCE_DEFAULT},
						  *timedOutAnchorP = &timedOutAnchor;
	msgIdIndexElement_t msgIdIndexAnchor = {MESSAGE_ID_INDEX_DEFAULT};
    logTruncationRequest_t logTruncationRequestAnchor = {LOG_TRUNCATION_REQUEST_DEFAULT},
		                   *logTruncationRequestAnchorP = &logTruncationRequestAnchor;
	int numLogTruncationRequests = 0;
	char *alarmThreadErrorMessage = NULL,
	     *logThreadErrorMessage = NULL;
	threadHandle_t alarmThreadHandle,
		           logThreadHandle;
	
	char queueManagerName[MQ_Q_MGR_NAME_LENGTH+1] = "";
	char brokerName[MAX_STRING_SIZE+1] = "";
	char executionGroupName[MAX_STRING_SIZE+1] = "";

	CciChar queueManagerNameUC[MQ_Q_MGR_NAME_LENGTH+1];
	CciChar brokerNameUC[MAX_STRING_SIZE+1];
	CciChar executionGroupNameUC[MAX_STRING_SIZE+1];

	/* trace stuff */
	char *callersTraceEnvironmentVariable = "AGGREGATOR_PLUGIN_TRACE";
	int traceCallerIndex;
	threadKey_t loggingKey = 0;

#else
extern addressList_t *addrListP;
	/* global variables */
    extern threadMutex_t aggregateChainMutex,
				         alarmThreadSyncMutex,
				         logThreadSyncMutex,
						 logThreadWorkMutex,
	  			         uuidGenMutex;
	extern threadCondVar_t alarmThreadWorkCondVar,
		                   alarmThreadSyncCondVar,
		                   logThreadWorkCondVar,
		                   logThreadSyncCondVar;
	extern bool_t shutdownRequested,
		          alarmThreadInitialised,
		          alarmThreadCreated,
                  logThreadInitialised,
                  logThreadCreated,
		          brokerInfoObtained,
		          multipleExecutionGroups,
				  alarmThreadShuttingDown,
				  logThreadShuttingDown,
		          firstTraceCall;
	extern messageFlow_t messageFlowAnchor;
    extern deployedAggregate_t deployedAggregateAnchor,
		                      *deployedAggregateAnchorP;
	extern aggregateInstance_t alarmAnchor,
	                             timedOutAnchor,
						         *timedOutAnchorP;
	extern msgIdIndexElement_t msgIdIndexAnchor;
    extern logTruncationRequest_t logTruncationRequestAnchor,
		                          *logTruncationRequestAnchorP;
	extern int numLogTruncationRequests;
	extern char *alarmThreadErrorMessage,
	            *logThreadErrorMessage;
	extern threadHandle_t alarmThreadHandle,
		                  logThreadHandle;

	extern char queueManagerName[];
	extern char brokerName[];
	extern char executionGroupName[];

	extern CciChar queueManagerNameUC[];
	extern CciChar brokerNameUC[];
	extern CciChar executionGroupNameUC[];
	
	/* trace stuff */
	extern char *callersTraceEnvironmentVariable;
	extern int traceCallerIndex;
	extern threadKey_t loggingKey;
#endif

/* add to double-linked list after current element */
#define INSERT_AFTER_IN_LINKED_LIST(__currentElementP, __newElementP) \
__newElementP->nextP = __currentElementP->nextP; \
__newElementP->prevP = __currentElementP; \
if (__currentElementP->nextP) __currentElementP->nextP->prevP = __newElementP; \
__currentElementP->nextP = __newElementP;

#define INSERT_AFTER_IN_ALARM_CHAIN(__currentElementP, __newElementP) \
__newElementP->nextAlarmP = __currentElementP->nextAlarmP; \
__newElementP->prevAlarmP = __currentElementP; \
if (__currentElementP->nextAlarmP) __currentElementP->nextAlarmP->prevAlarmP = __newElementP; \
__currentElementP->nextAlarmP = __newElementP; 

/* remove from double-linked list */
#define REMOVE_FROM_LINKED_LIST(__currentElementP) \
__currentElementP->prevP->nextP = __currentElementP->nextP; \
if (__currentElementP->nextP) \
	__currentElementP->nextP->prevP = __currentElementP->prevP;

#define REMOVE_FROM_ALARM_CHAIN(__currentElementP) \
__currentElementP->prevAlarmP->nextAlarmP = __currentElementP->nextAlarmP; \
if (__currentElementP->nextAlarmP) \
	__currentElementP->nextAlarmP->prevAlarmP = __currentElementP->prevAlarmP;


/* Function Prototypes */
int evaluateAggregateCreate(context_t *contextP,
		                   CciMessage *destinationList,
		                   CciMessage *exceptionList,
		                   CciMessage *message);

int evaluateAggregateReply(context_t *contextP,
		                   CciMessage *destinationList,
		                   CciMessage *exceptionList,
		                   CciMessage *message);

void setAggregateReplyAlarm(IN aggregateInstance_t *currentAggregateP,
	  	                    IN aggregateMember_t *memberP,
							IN bool_t recoveryProcessing);

int processAttribute(IN context_t *contextP,
				     IN CciChar *attributeName,
				     IN CciChar *attributeValue);

deployedAggregate_t *findDeployedAggregate(IN CciChar *name);

aggregateInstance_t *findAggregateInstanceFromContext(IN context_t *contextP,
														IN char *aggregateId);

void freeAllAggregateMembers(IN aggregateInstance_t *aggregateP);

threadFunction_t processAlarms(void *dataP);

threadFunction_t truncateLog(void *dataP);

void updateAlarmPairsTable(void *contextP,
						   CciChar *alarmName);

int notifyAggregateReplyNodeOfExpiredAggregate(IN MQHCONN qmConn,
											   IN aggregateInstance_t *aggregateP);

int buildCompletedAggregateReplyMessage(IN CciMessage *message,
						                IN aggregateInstance_t *aggregateP,
							            OUT CciMessage **newMessageP);

int processExpiredAggregate(IN context_t *contextP,
						    IN aggregateInstance_t *aggregateP,
                            IN CciMessage *message,
		   		            IN CciMessage *destinationList,
	                        IN CciMessage *exceptionList);

int buildAggregateMessagePayload(IN aggregateInstance_t *aggregateP,
								 OUT char **bufferP,
								 OUT int *lengthP,
								 OUT int *numberOfRepliesP);

int addAddressToDestinationList(IN CciMessage *destinationList,
								IN CciChar *queueName,
								IN CciChar *queueManagerName,
								OUT CciMessage **newDestinationListP);

msgIdIndexElement_t *updateMsgIdIndex(IN char *messageId);

void removeMsgIdIndex(IN msgIdIndexElement_t *msgIndexP);

msgIdIndexElement_t *findMsgIndexElement(IN char *correlId);

int createAlarmThread(IN context_t *contextP);

int copyMessage(IN CciMessage *newMsg, 
			    IN CciMessage *oldMsg);

CciMessage *createNewMessage(IN CciMessage *message);

void *getMessageIdElement(IN void *message);

void *findChild(IN void *currentElementP,
				IN CciChar *elementName);

void *getMessageElement(IN void *message,
					    IN matchType_t matchType);

int getMessageCharArrayValue(IN void *message,
							 IN matchType_t matchType,
							 OUT CciChar *buffer,
							 INOUT int *bufferLengthP,
							 OUT void **elementPP);

int swapMessageReplyDetails(IN CciMessage *message, 
							IN CciChar *newReplyToQueue,
							IN CciChar *newReplyToQueueManager,
							OUT CciChar *oldReplyToQueue,
							IN int oldReplyToQueueLength,
							OUT CciChar *oldReplyToQueueManager,
							IN int oldReplyToQueueManagerLength);

int swapMessageIdAndCorrelId(INOUT CciMessage *message);

void addMemberToAggregateInstance(IN aggregateMember_t *memberP,
		                          IN aggregateInstance_t *aggregateP);

void *replaceMessageCorrelId(IN void *message, 
							 IN char *correlId);

int propagateMessage(IN context_t *contextP,
					 IN CciChar *terminalName,
					 IN void *message,
					 IN void *destinationList,
					 IN void *exceptionList,
                     IN bool_t newMessage);	

#if 0
void deleteAggregateInstances(void);
#else
void deleteAggregateInstances(IN context_t *contextP,
							  IN deployedAggregate_t *deployedAggregateP);
#endif

int processTimerMessage(IN context_t *contextP,
 			   	        IN CciMessage *destinationList,
		                IN CciMessage *exceptionList,
		                IN CciMessage *message);

int convertToUnicode(IN char *inputString,
					 IN int inputStringLength,
					 OUT CciChar *outputString,
					 IN int outputStringLength);

int deleteMessageBody(INOUT CciMessage *message);

int replaceMessageBodyWithBLOB(INOUT CciMessage *message,
						       IN char *payload,
							   IN int payloadLength);

int writeAggregateInstanceToDB(IN context_t* contextP,
							   IN aggregateInstance_t *aggregateInstanceP,
 			   		           IN CciMessage *destinationList,
		                       IN CciMessage *exceptionList,
		                       IN CciMessage *message);

int writeMsgIdToDB(IN context_t* contextP,
				   IN char *msgId,
 		 	   	   IN CciMessage *destinationList,
		           IN CciMessage *exceptionList,
		           IN CciMessage *message);

int readMsgIdFromDB(IN context_t* contextP,
				    OUT char *msgId,
 		 	   	    IN CciMessage *destinationList,
		            IN CciMessage *exceptionList,
		            IN CciMessage *message);

void dispHex(IN char *inputString, 
			 IN int inputStringlength, 
			 OUT char *hexString);

void readMsgData(IN CciMessage *msg);

void traverseMessage(IN void* element, 
					 IN int blankCount,
					 IN int childLevel);

messageFlow_t *findMessageFlow(IN CciChar *name);

/* Node Implementation Functions */
void *createNodeContext(void    *factoryObject,
                        CciChar *nodeName,
                        void    *nodeObject);

int setAttribute(void    *context,
                 CciChar *attributeName,
                 CciChar *attributeValue);

void evaluate(void *context, 
              void *destinationList,
              void *exceptionList,
              void *message);

int genMessageId(OUT char *msgId);

void runTheChain();
void showGlobals(IN deployedAggregate_t *deployedAggregateP);
void traceMessageBuffer(IN CciMessage *message);
int getMessageLength(IN CciMessage *message,
					 OUT int *lengthP);
#if 0
void testAddress(int label);
#endif

#endif

